# coding: utf-8
from typing import List

from django.db.models import CharField
from django.db.models import ForeignKey
from django.db.models import IntegerField
from django.db.models import Model
from django.db.models import Q
from django.db.models import QuerySet
from django.db.models import TextField


class BaseServiceProvider:
    """Базовый провайдер для работы с данными."""

    def __init__(
        self,
        id_param_name: str = "id",
    ):
        self.id_param = id_param_name
        self._model = None

    def init(self, model: Model):
        """Инициализирует провайдер."""
        self._model = model

    @property
    def model(self):
        """Модель для доступа к данным."""
        return self._model

    def get_obj_by_id(self, pk: int):
        """Возвращает экземпляр модели по id."""
        obj = self.model.objects.filter(id=pk).first()
        return obj

    def create_obj(self, data: dict):
        """Создание экземпляра модели."""
        obj = self.model(**data)
        obj.full_clean()
        obj.save()
        return obj

    def update_obj(self, pk: int, data: dict):
        """Обновление экземпляра модели по заданной data."""
        obj = self.get_obj_by_id(pk)
        for attr, value in data.items():
            setattr(obj, attr, value)
        obj.full_clean()
        obj.save()
        return obj

    def get_all_model_objects(self) -> QuerySet:
        """Возвращает все экземпляры модели."""
        return self.model.objects.all()

    @staticmethod
    def get_search_filter_by_fields(search_fields_data: dict) -> Q:
        """Возвращает фильтр для поиска/"""
        filter_ = Q()
        for field, search_text in search_fields_data.items():
            filter_word = Q()
            filter_word |= Q(**{field: search_text})
            filter_ &= filter_word
        return filter_

    def get_search_f_filter(self, filters) -> Q:
        """Возвращает фильтр для поиска по f параметрам запроса."""
        filter_keys_checker = {}
        main_filter = Q()
        for element in filters:
            k, v = element.split('=')
            v = v.split(',')

            filter_ = Q()
            model_field = self.model._meta.get_field(k)
            if isinstance(model_field, (IntegerField, ForeignKey)):
                filter_ = Q(**{f'{k}__in': v})

            elif isinstance(model_field, (CharField, TextField)):
                filter_ = Q(**{f'{k}__icontains': v[0]})

            if filter_keys_checker.get(k):
                main_filter |= filter_
            else:
                filter_keys_checker[k] = 'key_exist'
                main_filter &= filter_
        return main_filter

    def delete_obj(self, pk: int):
        """Удаление объекта."""
        obj = self.get_obj_by_id(pk)
        obj.safe_delete()

    def get_objects_by_search_fields(
            self,
            limit,
            offset,
            search_fields_data: dict = None,
    ) -> QuerySet:
        """Возвращает список объектов модели."""
        query = self.get_all_model_objects()
        if search_fields_data:
            search_filter = self.get_search_filter_by_fields(
                search_fields_data)
            if search_filter:
                query = query.filter(search_filter)
        query = query[offset:offset + limit]
        return query

    def get_objects_by_ids(self, ids: List[int]) -> QuerySet:
        """Возвращает список моделей по списку id."""
        return self.model.objects.filter(id__in=ids)

    def get_objects(self, limit, offset, ids=None, filters=None) -> QuerySet:
        """Возвращает список объектов модели."""
        if ids and isinstance(ids, (List, tuple)):
            query = self.get_objects_by_ids(ids)
        else:
            query = self.get_all_model_objects()

        if filters:
            search_filter = self.get_search_f_filter(filters)
            if search_filter:
                query = query.filter(search_filter)
        query = query[offset:offset + limit]
        return query
